package com.example.android.apis.graphics;

import android.util.Log;
import android.util.SparseArray;
import android.view.InputDevice;

// early OUYA controllers seem to have some issues with analog stick accuracy.
// in what appears to be a bit of bad firmware, there are distinct bands of 'empty' space
// in the readings which makes values jump around quite and generally just wreaks havoc on
// the feel of the controller.

// see:  http://www.files.froemling.net/images/joystick_test/js_ouya.png
// vs an XBox360: http://www.files.froemling.net/images/joystick_test/js_xbox.png

// ...so to combat this, here is a filter which will group continuous sets of input data into
// 'buckets' and remap those buckets to 'fill in' the empty space.  This doesn't entirely
// fix the problem (areas near the axes will still always map to zero) but it definitely helps.

// This should be pretty harmless on non-broken controllers since in that case
// it should essentially detect one big continuous 'bucket' so remapping will be minimal or nonexistant.
// ..it still might be nice to try to detect 'fixed' controllers in the future just to avoid
// this computation though.  The filter currently ignores anything not called 'OUYA Game Controller'
// for that same reason.

// to use this, just create a single instance of it and run all your axis values through
// getFilteredAxisValue().  It will calibrate itself on the fly and give you filtered values.

class OuyaControllerFilter{
	
	public float getFilteredAxisValue(InputDevice device, int axis, float rawValue) {
		return getFilter(device,axis,rawValue).filter(rawValue);
	}
	
	// array of filters indexed by device-id and sub-indexed by axis+positive/negative-ness
	private SparseArray<SparseArray<Filter>> mFilters = new SparseArray<SparseArray<Filter>>();

	private Filter getFilter(InputDevice device, int axis,float rawValue){
		SparseArray<Filter> deviceFilters = mFilters.get(device.getId());
		if (deviceFilters == null){
			deviceFilters = new SparseArray<Filter>();
			mFilters.put(device.getId(), deviceFilters);
		}
		// keep a separate filter for positive/negative values of each axis
		int index = axis*2 + ((rawValue > 0.0f) ? 1:0);
		Filter f = deviceFilters.get(index);
		if (f == null){
			f = new Filter(device);
			deviceFilters.put(index,f);
		}
		return f;
	}
	private class Filter{
		
		// keep track of the smallest non-zero value we get so we can map that to zero
		private float mMinFinalVal = 0.35f;
		
		// a bucket is a continuous range of values that we're getting from the controller
		// we remap buckets to fill in the spaces between buckets
		private class Bucket{
			Bucket(float initialValue){
				mMinVal = mMaxVal = initialValue;
			}
			float mMinVal;
			float mMaxVal;
			Bucket next;
		}
		private Bucket mBuckets = null;
		
		Filter(InputDevice device){
			// we only want to run filtering on OUYA controllers.
			// (though technically it shouldn't hurt to filter anything; its just wasteful)
			mIsOuyaController = (device.getName().equals("OUYA Game Controller"));
		}
		float filter(float rawValue){
			if (!mIsOuyaController) return rawValue;
			
			float sign = (rawValue > 0.0f)?1.0f:-1.0f;
			float value = rawValue * sign; // always work in positive
			
			// find or create a bucket for this value
			Bucket b = mBuckets;
			Bucket bPrev = null;
			boolean changed = false;
			while (true){
				// if there's no more buckets, tack one on the end
				if (b == null){
					Bucket bNew = new Bucket(value);
					if (bPrev != null) bPrev.next = bNew;
					else mBuckets = bNew;
					changed = true;
					break;
				}
				// if we're before the current bucket, drop one in
				if (value < b.mMinVal){
					Bucket bNew = new Bucket(value);
					if (bPrev != null) bPrev.next = bNew;
					else mBuckets = bNew;
					bNew.next = b;
					changed = true;
					break;
				}
				// if we're *in* the current bucket, we're done
				if (value >= b.mMinVal && value <= b.mMaxVal){
					break;
				}
				// otherwise move to the next bucket
				bPrev = b;
				b = b.next;
			}
			
			// if anything's changed, go through our bucket list combining buckets that are close together
			if (changed){
				b = mBuckets;
				while (b != null){
					if (b.next != null && b.next.mMinVal - b.mMaxVal < 0.05){
						b.mMaxVal = b.next.mMaxVal;
						b.next = b.next.next;
					}
					else b = b.next;
				}
				// re-find the bucket containing our point since it might have been recombined
				b = mBuckets;
				while (b != null){
					if (b.mMaxVal >= value){
						break;
					}
					b = b.next;
				}
			}
			
			int bNum = 0;
			Bucket bTest = mBuckets;
			while (bTest != null){
				bNum++;
				Log.v("OUYA","span "+bNum+": "+bTest.mMinVal+" to "+bTest.mMaxVal);
				bTest = bTest.next;
			}

			// now remap the space in our bucket out to the beginning of the next bucket
			// (if there is a next bucket)
			if (b.next != null){
				// if our sample is a single point or is close to zero just go with the low value
				if (b.mMaxVal == b.mMinVal || b.mMaxVal < 0.001f) value = b.mMinVal;
				else value = b.mMinVal + (value-b.mMinVal)*(b.next.mMinVal-b.mMinVal)/(b.mMaxVal-b.mMinVal);
			}

			// lastly we want to remap to fill in the dead-zones near the axes
			// update: disabling this for now; I prefer just leaving those zones dead
			// rather than biasing speeds.
			
			// (you'll still have to pull the stick away from the axis significantly but at least small
			// values will be possible to achieve)
			// ...to do this we keep track of the minimum non-zero value we've gotten and stretch everything back so that maps to zero
//			if (value > 0.001){
//				if (value < mMinFinalVal) mMinFinalVal = value;
//				// now stretch the value out from 0.5 so that our min known value maps to 0.0
//				// (this will probably send our high values past 1.0 so we clamp there too..)
//				value = 0.5f + (value-0.5f) * (0.0f-0.5f)/(mMinFinalVal-0.5f);
//				if (value > 1.0f) value = 1.0f;
//			}
			return value * sign;
		}
		private boolean mIsOuyaController;
	}
}
